#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-------------------------------------------------------------------------------
# Use of this source is governed by the MIT license, see LICENSE file.
#-------------------------------------------------------------------------------


"""
Base library for web interface unit tests.
"""


import io
import unittest
import pprint
import logging
import flask_mail
import vial.db


#logging.disable(logging.CRITICAL+1000)


DEMO_GROUP_A = 'DEMO_GROUP_A'
DEMO_GROUP_B = 'DEMO_GROUP_B'


def _config_testapp_vial(app_config, config_func):
    """
    Configure and reconfigure application for testing before instantination.
    """
    config_func(app_config)
    app_config['TESTING'] = True
    app_config['WTF_CSRF_ENABLED'] = False
    app_config['DEBUG'] = False
    app_config['EXPLAIN_TEMPLATE_LOADING'] = False
    app_config['LOG_FILE'] = '/var/tmp/vial-utest.log'
    app_config['SQLALCHEMY_DATABASE_URI'] = 'postgresql://vial:vial@localhost/vial_utest'
    app_config['MAIL_SERVER'] = 'localhost'
    app_config['MAIL_PORT'] = 8025
    app_config['MAIL_DEFAULT_SENDER'] = 'root@unittest'
    app_config['EMAIL_ADMINS'] = ['admin@unittest']


class do_as_user_decorator:  # pylint: disable=locally-disabled,invalid-name
    """
    Decorator class for accessing application endpoints as given user.
    """
    def __init__(self, user_name):
        self.user_name = user_name

    def __call__(self, func):
        def wrapped_f(other_self, *args, **kwargs):
            response = other_self.login_dev(self.user_name)
            other_self.assertEqual(response.status_code, 200)

            func(other_self, *args, **kwargs)

            response = other_self.logout()
            other_self.assertEqual(response.status_code, 200)

        return wrapped_f


class LoggedTestCaseMetaclass(type):
    def __new__(mcs, name, bases, dct):
        # if the TestCase already provides setUp, wrap it
        if 'setUp' in dct:
            setUp = dct['setUp']  # pylint: disable=locally-disabled,invalid-name
        else:
            setUp = lambda self: None  # pylint: disable=locally-disabled,invalid-name

        def wrappedSetUp(self):  # pylint: disable=locally-disabled,invalid-name
            for hdlr in self.logger.handlers:
                self.logger.removeHandler(hdlr)
            self.loghdlr = logging.StreamHandler(io.StringIO())
            self.logger.addHandler(self.loghdlr)
            setUp(self)
        dct['setUp'] = wrappedSetUp

        # same for tearDown
        if 'tearDown' in dct:
            tearDown = dct['tearDown']  # pylint: disable=locally-disabled,invalid-name
        else:
            tearDown = lambda self: None  # pylint: disable=locally-disabled,invalid-name

        def wrappedTearDown(self):  # pylint: disable=locally-disabled,invalid-name
            tearDown(self)
            self.loghdlr.stream.close()
            self.logger.removeHandler(self.loghdlr)
        dct['tearDown'] = wrappedTearDown

        # return the class instance with the replaced setUp/tearDown
        return type.__new__(mcs, name, bases, dct)


class VialTestCase(unittest.TestCase):
    """
    Class for testing :py:class:`vial.app.Vial` application.
    """
    #__metaclass__ = LoggedTestCaseMetaclass

    logger = logging.getLogger()
    logger.level = logging.DEBUG

    def on_email_sent(self, message, app):
        """
        Signal handler for handling :py:func:`flask_mail.email_dispatched` signal.
        Log subject and recipients of all email that have been sent.
        """
        #pprint.pprint(message)
        #app.logger.info(
        #    "Sent email '%s' to '%s'",
        #    message.subject,
        #    ', '.join(message.recipients)
        #)
        self.mailbox.append(message)

    def setUp(self):
        self.setup_logging()
        self.mailbox = []
        self.fixtures_db = []

        self.app = self.setup_app()
        self.client = self.app.test_client()
        self.setup_db()

    def tearDown(self):
        self.teardown_logging()

    def setup_logging(self):
        #for hdlr in self.logger.handlers:
        #    self.logger.removeHandler(hdlr)
        self.loghdlr = logging.StreamHandler(io.StringIO())
        self.logger.addHandler(self.loghdlr)

    def setup_app(self):
        """
        Setup application object.
        """
        raise NotImplementedError()

    def setup_db(self):
        """
        Perform database setup.
        """
        with self.app.app_context():
            vial.db.db_get().drop_all()
            vial.db.db_get().create_all()
            self.setup_fixtures_db()

    def setup_fixtures_db(self):
        """
        Setup database object fixtures.
        """
        user_model  = self.app.get_model(vial.const.MODEL_USER)
        group_model = self.app.get_model(vial.const.MODEL_GROUP)

        def _gen_user(user_type):
            return user_model(
                login = user_type,
                fullname = 'Demo {}'.format(user_type[0].upper() + user_type[1:]),
                email = '{}@bogus-domain.org'.format(user_type),
                organization = 'BOGUS DOMAIN, a.l.e.',
                roles = list(set([vial.const.ROLE_USER, user_type])),
                enabled = True,
                apikey = 'apikey-{}'.format(user_type)
            )

        account_user = _gen_user(vial.const.ROLE_USER)
        account_developer = _gen_user(vial.const.ROLE_DEVELOPER)
        account_maintainer = _gen_user(vial.const.ROLE_MAINTAINER)
        account_admin = _gen_user(vial.const.ROLE_ADMIN)

        group_a = group_model(
            name = DEMO_GROUP_A,
            source = 'manual',
            description = 'Demo Group A',
            enabled = True
        )
        group_b = group_model(
            name = DEMO_GROUP_B,
            source = 'manual',
            description = 'Demo Group B',
            enabled = True
        )
        group_a.members.append(account_user)
        group_a.members.append(account_developer)
        group_a.members.append(account_maintainer)
        group_a.members.append(account_admin)

        group_a.managers.append(account_developer)
        group_a.managers.append(account_maintainer)

        fixture_list = [account_user, account_developer, account_maintainer, account_admin, group_a, group_b]
        for dbobject in fixture_list:
            vial.db.db_session().add(dbobject)
            vial.db.db_session().commit()

        self.fixtures_db.extend(fixture_list)

    def teardown_logging(self):
        #print(
        #    "CAPTURED LOG CONTENTS:\n{}".format(
        #        self.loghdlr.stream.getvalue()
        #    )
        #)
        self.loghdlr.stream.close()
        self.logger.removeHandler(self.loghdlr)

    def login_dev(self, login):
        """
        Login given user with *auth_dev* module.
        """
        return self.client.post(
            '/auth_dev/login',
            data = dict(login = login, submit = 'Login'),
            follow_redirects = True
        )

    def login_pwd(self, login, password):
        """
        Login given user with *auth_pwd* module.
        """
        return self.client.post(
            '/auth_pwd/login',
            data = dict(login = login, password = password, submit = 'Login'),
            follow_redirects = True
        )

    def login_env(self, login, envvar = 'eppn'):
        """
        Login given user with *auth_env* module.
        """
        return self.client.get(
            '/auth_env/login',
            environ_base = {envvar: login},
            follow_redirects = True
        )

    def logout(self):
        """
        Logout current user.
        """
        return self.client.get(
            '/logout',
            follow_redirects = True
        )

    def log_get(self):
        """
        Get content written to log so far.
        """
        return self.loghdlr.stream.getvalue()

    def log_clear(self):
        """
        Clear log content.
        """
        self.loghdlr.stream.close()
        self.loghdlr.stream = io.StringIO()

    def mailbox_clear(self):
        """
        Clear internal mailbox.
        """
        self.mailbox = []

    def mailbox_monitoring(self, state):
        """
        Enable/disable mailbox monitoring.
        """
        if state == 'on':
            flask_mail.email_dispatched.connect(self.on_email_sent)
            return

        if state == 'off':
            flask_mail.email_dispatched.disconnect(self.on_email_sent)
            return

        raise ValueError(
            'Invalid parameter for mailbox_monitoring, must be "on" or "off", received {}'.format(
                str(state)
            )
        )

    def user_get(self, user_type, with_app_context = False):
        """
        Get given user.
        """
        if not with_app_context:
            user_model = self.app.get_model(vial.const.MODEL_USER)
            return vial.db.db_session().query(user_model).filter(user_model.login == user_type).one_or_none()
        with self.app.app_context():
            user_model = self.app.get_model(vial.const.MODEL_USER)
            return vial.db.db_session().query(user_model).filter(user_model.login == user_type).one_or_none()

    def user_save(self, user_object, with_app_context = False):
        """
        Update given user.
        """
        if not with_app_context:
            vial.db.db_session().add(user_object)
            vial.db.db_session().commit()
        with self.app.app_context():
            vial.db.db_session().add(user_object)
            vial.db.db_session().commit()

    def user_id(self, user_type, with_app_context = False):
        """
        Get ID of given user.
        """
        if not with_app_context:
            uobj = self.user_get(user_type)
            return uobj.id
        with self.app.app_context():
            uobj = self.user_get(user_type)
            return uobj.id

    def user_enabled(self, user_type, state, with_app_context = False):
        """
        Enable/disable given user.
        """
        if not with_app_context:
            user = self.user_get(user_type)
            user.enabled = state
            self.user_save(user)
            return
        with self.app.app_context():
            user = self.user_get(user_type, False)
            user.enabled = state
            self.user_save(user, False)
            return

    def group_get(self, group_name, with_app_context = False):
        """
        Get given group.
        """
        if not with_app_context:
            group_model = self.app.get_model(vial.const.MODEL_GROUP)
            return vial.db.db_session().query(group_model).filter(group_model.name == group_name).one_or_none()
        with self.app.app_context():
            group_model = self.app.get_model(vial.const.MODEL_GROUP)
            return vial.db.db_session().query(group_model).filter(group_model.name == group_name).one_or_none()

    def group_save(self, group_object, with_app_context = False):
        """
        Update given group.
        """
        if not with_app_context:
            vial.db.db_session().add(group_object)
            vial.db.db_session().commit()
        with self.app.app_context():
            vial.db.db_session().add(group_object)
            vial.db.db_session().commit()

    def group_id(self, group_name, with_app_context = False):
        """
        Get ID of given group.
        """
        if not with_app_context:
            gobj = self.group_get(group_name)
            return gobj.id
        with self.app.app_context():
            gobj = self.group_get(group_name)
            return gobj.id

    def assertGetURL(self, url, status_code = 200, content_checks = None, print_response = False):  # pylint: disable=locally-disabled,invalid-name
        """
        Perform GET request and check some default assertions against the response.
        """
        response = self.client.get(
            url,
            follow_redirects = True
        )
        if print_response:
            print("--------------------------------------------------------------------------------")
            print("Response for {}: {} ({})".format(url, response.status_code, response.status))
            pprint.pprint(response.headers)
            pprint.pprint(response.data)
            print("--------------------------------------------------------------------------------")
        self.assertEqual(response.status_code, status_code)
        if content_checks:
            for cch in content_checks:
                self.assertTrue(cch in response.data)
        return response

    def assertPostURL(self, url, data, status_code = 200, content_checks = None, print_response = False):  # pylint: disable=locally-disabled,invalid-name
        """
        Perform POST request and check some default assertions against the response.
        """
        response = self.client.post(
            url,
            data = data,
            follow_redirects = True
        )
        if print_response:
            print("--------------------------------------------------------------------------------")
            print("Response for {}, {}: {} ({})".format(url, pprint.pformat(data), response.status_code, response.status))
            pprint.pprint(response.headers)
            pprint.pprint(response.data)
            print("--------------------------------------------------------------------------------")
        self.assertEqual(response.status_code, status_code)
        if content_checks:
            for cch in content_checks:
                self.assertTrue(cch in response.data)
        return response

    def assertMailbox(self, checklist):  # pylint: disable=locally-disabled,invalid-name
        """
        Check internal mailbox.
        """
        for attr_name in ('subject', 'sender', 'recipients', 'cc', 'bcc', 'body', 'html'):
            if attr_name in checklist:
                self.assertEqual(
                    list(
                        map(
                            lambda x: getattr(x, attr_name),
                            self.mailbox,
                        )
                    ),
                    checklist[attr_name]
                )

class ItemCreateVialTestCase(VialTestCase):
    """
    Class for testing :py:class:`vial.app.Vial` application item creation views.
    """
    maxDiff = None

    def assertCreate(self, url, data, content_checks = None, print_response = False):  # pylint: disable=locally-disabled,invalid-name
        response = self.client.get(
            url,
            follow_redirects = True
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'<div class="btn-toolbar" role="toolbar" aria-label="Form submission buttons">' in response.data)

        #
        # Disable this feature for now. TODO: Rethink and redesign.
        #
        #for idx, param in enumerate(data):
        #    if idx == len(data) - 1:
        #        break
        #    response = self.client.post(
        #        url,
        #        follow_redirects = True,
        #        data = {
        #            i[0]: i[1] for i in data[0:idx+1]
        #        }
        #    )
        #    self.assertEqual(response.status_code, 200)
        #    self.assertTrue(b'This field is required.' in response.data)
        #    self.assertTrue(b'help-block form-error' in response.data)

        response = response = self.client.post(
            url,
            follow_redirects = True,
            data = {
                i[0]: i[1] for i in data
            }
        )
        if print_response:
            print("--------------------------------------------------------------------------------")
            print("Response for {}, {}: {} ({})".format(url, pprint.pformat(data), response.status_code, response.status))
            pprint.pprint(response.headers)
            pprint.pprint(response.data)
            print("--------------------------------------------------------------------------------")
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'<div class="alert alert-success alert-dismissible">' in response.data)
        if content_checks:
            for cch in content_checks:
                self.assertTrue(cch in response.data)
        return response


class RegistrationVialTestCase(VialTestCase):
    """
    Class for testing :py:class:`vial.app.Vial` application registration views.
    """
    maxDiff = None

    def assertRegisterFail(self, url, data, environ_base = None):  # pylint: disable=locally-disabled,invalid-name
        response = response = self.client.get(
            url,
            follow_redirects = True,
            environ_base = environ_base,
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'User account registration' in response.data)

        for idx, param in enumerate(data):
            if idx == len(data) - 1:
                break
            response = response = self.client.post(
                url,
                follow_redirects = True,
                environ_base = environ_base,
                data = {
                    i[0]: i[1] for i in data[0:idx+1]
                }
            )
            self.assertEqual(response.status_code, 200)
            self.assertTrue(b'This field is required.' in response.data)
            self.assertTrue(b'help-block form-error' in response.data)

        response = response = self.client.post(
            url,
            follow_redirects = True,
            environ_base = environ_base,
            data = {
                i[0]: i[1] for i in data
            }
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            b'Please use different login, the &#34;user&#34; is already taken.' in response.data or \
            b'Please use different login, the "user" is already taken.' in response.data
        )


    def assertRegister(self, url, data, emails, environ_base = None):  # pylint: disable=locally-disabled,invalid-name
        uname = 'test'
        self.mailbox_monitoring('on')

        response = response = self.client.get(
            url,
            follow_redirects = True,
            environ_base = environ_base
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'User account registration' in response.data)

        for idx, param in enumerate(data):
            if idx == len(data) - 1:
                break
            response = response = self.client.post(
                url,
                follow_redirects = True,
                environ_base = environ_base,
                data = {
                    i[0]: i[1] for i in data[0:idx+1]
                }
            )
            self.assertEqual(response.status_code, 200)
            self.assertTrue(b'This field is required.' in response.data)
            self.assertTrue(b'help-block form-error' in response.data)

        response = response = self.client.post(
            url,
            follow_redirects = True,
            environ_base = environ_base,
            data = {
                i[0]: i[1] for i in data
            }
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'User account <strong>test (Test User)</strong> was successfully registered.' in response.data)
        with self.app.app_context():
            uobj = self.user_get(uname)
            self.assertTrue(uobj)
        self.assertMailbox(
            {
                'subject': [
                    '[{}] Account registration - {}'.format(self.app.config['APPLICATION_NAME'], uname),
                    '[{}] Account registration - {}'.format(self.app.config['APPLICATION_NAME'], uname)
                ],
                'sender': [
                    'root@unittest',
                    'root@unittest'
                ],
                'recipients': [
                    ['admin@unittest'],
                    ['test.user@domain.org']
                ],
                'cc': [[],[]],
                'bcc': [[], ['admin@unittest']],
                'body': emails['txt'],
                'html': emails['html']
            }
        )

        self.mailbox_monitoring('off')

        with self.app.app_context():
            user = self.user_get(uname)
            user_dict = user.to_dict()
            del user_dict['createtime']
            del user_dict['password']
            self.assertEqual(
                user_dict,
                {
                    'apikey': None,
                    'email': 'test.user@domain.org',
                    'enabled': False,
                    'fullname': 'Test User',
                    'id': 5,
                    'locale': None,
                    'login': 'test',
                    'logintime': 'None',
                    'managements': [],
                    'memberships': [],
                    'memberships_wanted': [],
                    'organization': 'TEST, org.',
                    'roles': ['user'],
                    'timezone': None
                }
            )
        response = self.login_dev(uname)
        self.assertEqual(response.status_code, 403)
        #self.assertTrue(b'is currently disabled, you are not permitted to log in.' in response.data)

        with self.app.app_context():
            user = self.user_get(uname)
            user.set_state_enabled()
            self.user_save(user)

        with self.app.app_context():
            user = self.user_get(uname)
            user_dict = user.to_dict()
            del user_dict['createtime']
            del user_dict['password']
            self.assertEqual(
                user_dict,
                {
                    'apikey': None,
                    'email': 'test.user@domain.org',
                    'enabled': True,
                    'fullname': 'Test User',
                    'id': 5,
                    'locale': None,
                    'login': 'test',
                    'logintime': 'None',
                    'managements': [],
                    'memberships': [],
                    'memberships_wanted': [],
                    'organization': 'TEST, org.',
                    'roles': ['user'],
                    'timezone': None
                }
            )
        response = self.login_dev(uname)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'You have been successfully logged in as' in response.data)

        response = self.logout()
        self.assertEqual(response.status_code, 200)
        self.assertTrue(b'You have been successfully logged out' in response.data)
